/*********************************************************************
 *                                                                   *
 *                     XGS : Apple IIGS Emulator                     *
 *                                                                   *
 *        Written and Copyright (C)1996 by Joshua M. Thompson        *
 *                                                                   *
 *  You are free to distribute this code for non-commercial purposes *
 * I ask only that you notify me of any changes you make to the code *
 *     Commercial use is prohibited without my written permission    *
 *                                                                   *
 *********************************************************************/

/*
 * File: emul.c
 *
 * Startup and shutdown code for the emulator, as well as the update
 * routine called 50 times per second.
 */

#include "xgs.h"
#include <sys/time.h>
#include <unistd.h>

#include "adb/adb.h"
#include "clock/clock.h"
#include "disks/disks.h"
#include "emul/emul.h"
#include "sound/sound.h"
#include "video/video.h"

int	emul_vbl;
int	emul_vblirq;
int	emul_qtrsecirq;
int	emul_onesecirq;

int	emul_cycles;
long	emul_delay;
int	emul_total_cycles;
double	emul_total_time;

double	emul_tick_times[60];
int	emul_tick_cycles[60];
int	emul_curr_tick;

double	emul_last_time;

int	emul_speed;
int	emul_speed2;

int EMUL_init()
{
	int	i,err;

	printf("%s\n",VERSION);
	printf("Written and (C) 1996 by Joshua M. Thompson\n");

	emul_vbl = 0;
	emul_vblirq = 0;
	emul_qtrsecirq = 0;
	emul_onesecirq = 0;
	emul_cycles = 20000;
	emul_delay = 0;
	emul_total_cycles = 0;
	emul_total_time = 0;
	emul_curr_tick = 0;
	emul_speed = 1;
	emul_speed2 = 0;

	for (i = 0 ; i < 60 ; i++) {
		emul_tick_times[i] = 0;
		emul_tick_cycles[0] = 0;
	}

	if ((err = MEM_init())) return err;
	if ((err = VID_init())) {
		MEM_shutdown();
		return err;
	}
	if ((err = SND_init())) {
		VID_shutdown();
		MEM_shutdown();
		return err;
	}
	if ((err = ADB_init())) {
		SND_shutdown();
		VID_shutdown();
		MEM_shutdown();
		return err;
	}
	if ((err = CLK_init())) {
		ADB_shutdown();
		SND_shutdown();
		VID_shutdown();
		MEM_shutdown();
		return err;
	}
	if ((err = DSK_init())) {
		CLK_shutdown();
		ADB_shutdown();
		SND_shutdown();
		VID_shutdown();
		MEM_shutdown();
		return err;
	}
	return 0;
}

void EMUL_update(int cycles)
{
	int		target_cycles;
	struct timeval	tv;
	double		this_time,diff;
	double		new_speed;
	char		buffer[256];

	usleep(emul_delay);
	gettimeofday(&tv,NULL);

	this_time = (double) tv.tv_sec + (double) tv.tv_usec / 1000000.0;
	diff = this_time - emul_last_time;

	emul_total_time = emul_total_time - emul_tick_times[emul_curr_tick] + diff;
	emul_tick_times[emul_curr_tick] = diff;

	emul_total_cycles = emul_total_cycles - emul_tick_cycles[emul_curr_tick] + cycles;
	emul_tick_cycles[emul_curr_tick] = cycles;

	new_speed = emul_total_cycles / emul_total_time;

/*
	emul_speed = 0;
*/

	if (!emul_speed) {
		target_cycles = 1024000;
	} else if (!emul_speed2) {
		target_cycles = 2560000;
	} else {
		target_cycles = 0;
	}

	if (emul_total_time < 0.998) {
		if (target_cycles) {
			emul_delay = (1.0 - emul_total_time) * 1000000 / 60.0;
			emul_cycles = target_cycles / 60;
		} else {
			emul_delay = 0;
			emul_cycles = new_speed / 60;
		}
	} else if (emul_total_time > 1.002) {
		emul_delay = 0;
		emul_cycles = new_speed / 60;
	}

	CPU_setUpdatePeriod(emul_cycles);

	emul_vbl ^= 0x80;
	if (emul_vbl == 0x00) {
		if (emul_vblirq) {
			mem_diagtype |= 0x08;
			CPU_irq();
		}
		MEM_update();
		VID_update();
		SND_update();
		ADB_update();
		CLK_update();
		DSK_update();
	}

	if (emul_qtrsecirq && (emul_curr_tick == 12)) {
		CPU_irq();
		mem_diagtype |= 0x10;
	}

	if (emul_onesecirq && (emul_curr_tick == 60)) {
		/* Not implemented yet */
	}

	emul_curr_tick++;
	if (emul_curr_tick == 60) {
		emul_curr_tick = 0;
		sprintf(buffer,"cycles = %5d, tcycles = %8d, time = %1.3f, ttime = %1.3f",
					cycles, emul_total_cycles, diff, emul_total_time);
		VID_drawStatus1(buffer);
		sprintf(buffer,"Target speed = %2.3f MHz, Actual speed = %2.3f MHz",
					target_cycles / 1000000.0, new_speed / 1000000.0);
		VID_drawStatus2(buffer);
	}

	emul_last_time = this_time;
}

void EMUL_reset()
{
	MEM_reset();
	VID_reset();
	SND_reset();
	ADB_reset();
	CLK_reset();
	DSK_reset();
	CPU_reset();
}

void EMUL_shutdown()
{
	DSK_shutdown();
	CLK_shutdown();
	ADB_shutdown();
	SND_shutdown();
	VID_shutdown();
	MEM_shutdown();
	exit(0);
}

void EMUL_trace(int parm)
{
	CPU_setTrace(parm);
}

void EMUL_handleWDM(byte parm)
{
	switch(parm) {
		case 0xC5 :	DSK_handleSlot5();
				break;
		case 0xC7 :	DSK_handleSlot7();
				break;
		case 0xFD :	CPU_setTrace(0);
				break;
		case 0xFE :	CPU_setTrace(1);
				break;
		case 0xFF :	EMUL_shutdown();
				break;
		default :	break;
	}
}

byte EMUL_getVBL(byte val)
{
	return emul_vbl;
}

void usage(char *myname)
{
	fprintf(stderr,"\nUsage: %s [-s5d1 filename ] [-s5d2 filename ] [-s6d1 filename ] [-s6d2 filename ] [-s7d1 filename ] [-s7d2 filename ] [-trace]\n",myname);
	exit(1);
}

void main(int argc, char *argv[])
{
	int		err,i;
	struct timeval	tv;

	i = 0;
	while (++i < argc) {
		if (!strcmp(argv[i],"-trace")) {
			EMUL_trace(1);
		} else if (!strcmp(argv[i],"-s5d1")) {
			if (++i >= argc) usage(argv[0]);
			s5d1_startup = argv[i];
		} else if (!strcmp(argv[i],"-s5d2")) {
			if (++i >= argc) usage(argv[0]);
			s5d2_startup = argv[i];
		} else if (!strcmp(argv[i],"-s6d1")) {
			if (++i >= argc) usage(argv[0]);
			s6d1_startup = argv[i];
		} else if (!strcmp(argv[i],"-s6d2")) {
			if (++i >= argc) usage(argv[0]);
			s6d2_startup = argv[i];
		} else if (!strcmp(argv[i],"-s7d1")) {
			if (++i >= argc) usage(argv[0]);
			s7d1_startup = argv[i];
		} else if (!strcmp(argv[i],"-s7d2")) {
			if (++i >= argc) usage(argv[0]);
			s7d2_startup = argv[i];
		} else usage(argv[0]);
	}
	if ((err = EMUL_init())) exit(err);
	printf("\nResetting the emulator...\n");
	EMUL_reset();
	printf("\nRunning!\n");

	gettimeofday(&tv,NULL);
	emul_last_time = (double) tv.tv_sec + (double) tv.tv_usec / 1000000.0;

	CPU_setUpdatePeriod(emul_cycles);
	CPU_run();
	EMUL_shutdown();
}
